"""adds cell magic functions for pidgy.

the cell magics we use are:
* pidgy - allows running pidgy tangle and weave without a persistent extension
* weave - weaves the cell into interactive ipython displays
* tangle - tangles the cell into python
"""
# this module needs work and thought about what the line portion of the magic does.
from . import get_ipython, get_cell_id


def tangle(line, cell):
    """tangle the cell to python code."""
    from IPython.display import Code

    return Code(get_ipython().tangle.render(cell), language="python")


def parse(line, cell):
    """tangle the cell to python code."""
    return get_ipython().tangle.parse(cell)


def weave(line, cell):
    """weave the cell input into a display"""
    shell = get_ipython()

    from IPython.core.interactiveshell import ExecutionInfo, ExecutionResult

    return shell.weave.post_run_cell(
        ExecutionResult(ExecutionInfo(cell, True, False, True, get_cell_id()))
    )


def pidgy(line, cell):
    """weave the cell input into a display"""
    shell = get_ipython()
    shell.run_cell(shell.tangle.render(cell), silent=True)
    return weave("", cell)


def load_ipython_extension(shell):
    from .tangle import _add_tangle_trait
    from .weave import _add_weave_trait, _add_interactivity

    _add_tangle_trait(shell)
    _add_weave_trait(shell)
    _add_interactivity(shell)

    # add filters as line magics for more reuse
    from . import filters

    for k, v in vars(filters).items():
        try:
            if v.__module__ == "filters.__name__":
                shell.register_magic_function(v, "line")
        except AttributeError:
            pass

    shell.register_magic_function(tangle, "cell")
    shell.register_magic_function(parse, "cell")
    shell.register_magic_function(weave, "cell")
    shell.register_magic_function(pidgy, "cell")


def unload_ipython_extension(shell):
    from .weave import _rm_interactivity

    _rm_interactivity(shell)



"""a modified pidgy kernel.

the command line application installs the kernel"""

from contextlib import suppress
from sys import argv

from ipykernel.ipkernel import IPythonKernel


class Kernel(IPythonKernel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.shell.magics_manager.magics["line"]["reload_ext"]("pidgy")



"""Reactive IPython/IPywidgets templates

these displays combine IPython displays with jinja2 environments.
we export multiple displays to use in different contexts.
for example, voila prefers HTML ipywidgets based displays to gain interactivity.
"""
# in this module we try to shield the library imports to make
# dependencies optional. we always have markdown it and ipython.
# ipywidgets may be optional

from dataclasses import dataclass, field
from io import StringIO
from re import compile
from IPython.display import DisplayHandle, display, HTML, Markdown

from markdown_it import MarkdownIt
from asyncio import ensure_future
from . import get_ipython

URL = compile("^(http[s]|file)://")


@dataclass
class TemplateDisplay:
    """the TemplateDisplays base class that provides the api for displaying and updating templates."""

    body: str = None
    template: object = None
    display_cls: type = Markdown
    display_handle: object = None
    iframe_attrs: dict = field(default_factory=dict(width="100%", height=600, loading="lazy").copy)
    is_list_urls: bool = None
    vars: set = field(default_factory=set)
    # this is not used by the base class but may be used by other base classes that render html
    tokens: object = None
    markdown_renderer: object = None
    use_async: bool = True

    def _ipython_display_(self):
        """display the template contents"""
        if self.display_handle is None:
            self.display_handle = DisplayHandle()

        if self.use_async and get_ipython().weave.reactive:
            self.display_handle.display(self.display_object(self.body))
            ensure_future(self.aupdate())
        else:
            try:
                import nest_asyncio

                nest_asyncio.apply()
            except ModuleNotFoundError:
                raise ImportError("nest_asyncio is needed to use synchronous rendering.")
            self.display_handle.display(self.display_object(self.render()))

    def _is_list_urls(self, x):
        for line in filter(bool, map(str.strip, StringIO(x))):
            if URL.match(line):
                continue

            return False
        return True

    async def arender(self):
        """async template rendering"""
        output = StringIO()

        try:
            async for part in self.template.generate_async():
                output.write(part)
        except BaseException as e:
            import traceback

            msg = "".join(traceback.format_exception(type(e), e, e.__traceback__))
            return f"""`````````pytb\n{msg}\n`````````"""
        return output.getvalue()

    def render(self):
        """sync template rendering"""
        return self.template.render()

    def display_object(self, object, **kwargs):
        return self.display_cls(object, **kwargs)

    async def aupdate(self):
        # it should be possible to do smarter updates
        if self.display_handle:
            self.display_handle.update(self.display_object(await self.arender()))

    def observe(self, _):
        ensure_future(self.aupdate())

    def is_widget(self):
        return is_widget(self.display_handle)

    def update(self):
        # it should be possible to do smarter updates
        if self.display_handle:
            self.display_handle.update(self.display_object(self.render()))

    def embed(self, urls):
        """we have a feature for showing iframes of urls.

        we can add richer features later like domain rules and file extension dispatchers
        maybe the could have been done with markdown it?"""
        lines = []
        # compose the default iframe attributes
        args = " ".join(f'{k}="{v}"' for k, v in self.iframe_attrs.items())

        # iterate through all the lines of the source and generate iframes
        # from them.
        for line in filter(bool, map(str.strip, StringIO(urls))):
            lines.append(f'<iframe src="{line}" {args}/>')
        return "\n".join(lines)


@dataclass
class IPythonMarkdown(TemplateDisplay):
    pass


@dataclass
class MarkdownItMixin:
    markdown_renderer: object = field(default_factory=MarkdownIt)

    async def arender(self):
        return self.markdown_renderer.render(await super().arender())


@dataclass
class IPythonHtml(MarkdownItMixin, IPythonMarkdown):
    display_cls: type = HTML


@dataclass
class IPyWidgetsHtml(MarkdownItMixin, TemplateDisplay):
    display_cls: "ipywidgets.Widget" = None

    def __post_init__(self):
        if self.display_cls is None:
            from ipywidgets import HTML

            self.display_cls = HTML

    async def aupdate(self):
        if self.display_handle:
            self.display_handle.value = await self.arender()

    def _ipython_display_(self):
        if self.display_handle is None:
            self.display_handle = self.display_object("")

        if self.use_async:
            display(self.display_handle)
            ensure_future(self.aupdate())
        else:
            display(self.display_object(self.render()))


def is_widget(object):
    from sys import modules

    if "ipywidgets" in modules:
        from ipywidgets import Widget

        return isinstance(object, Widget)
    return False



"""a modified pidgy kernel.

the command line application installs the kernel"""

from contextlib import suppress

from ipykernel.ipkernel import IPythonKernel


class Kernel(IPythonKernel):
    # help_links = List([]).tag(config=True)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shell.magics_manager.magics["line"]["reload_ext"]("pidgy")


if __name__ == "__main__":
    import ipykernel.kernelapp

    ipykernel.kernelapp.IPKernelApp.launch_instance(kernel_class=Kernel)



"jinja2 filters that make writing pidgy more fun"

from . import get_ipython


def q(q, cite):
    """<q>uotes"""
    return f"""<q cite="{cite}">{q}</q>"""


def i(id, height=600, width="100%"):
    """a filter for <q>uotes"""
    from urllib.parse import urlparse
    from html import escape

    parse = urlparse(id)

    src = "src"
    if parse.scheme and parse.netloc:
        id = escape(id)
        src += "doc"
    return f"""<iframe {src}="{id}" width="{width}" height="{height}"/>"""


def md(body):
    return get_ipython().weave.markdown_renderer.render(body)



import json
from pathlib import Path
import logging
from os import environ, system
from sys import executable
from subprocess import check_output
from shutil import copy

ROOT = Path(__file__).parent.parent
LOGO = ROOT / "pidgy.png"
DOCS = ROOT / "docs"

log = logging.getLogger("mkdocs")


def on_pre_build(config):
    if not (DOCS / LOGO.name).exists():
        log.info("move logo to docs dir")
        copy(LOGO, DOCS)


def on_post_build(config):
    if "READTHEDOCS" in environ:
        C = ROOT / "lite" / "jupyter_lite_config.json"
        data = json.loads(C.read_text())
        data["LiteBuildConfig"]["output_dir"] = str(Path(config.site_dir) / "run")
        C.write_text(json.dumps(data))
        log.info("building lite")
        check_output([executable, "-m", "doit", "lite"], cwd=ROOT)


# https://www.mkdocs.org/user-guide/configuration/#hooks



from . import get_ipython, get_cell_id
from dataclasses import dataclass
import ast, types, dis
from contextlib import suppress


@dataclass
class Execution:
    id: str
    tokens: list = None
    py: str = None
    nodes: ast.AST = None
    bytecode: types.CodeType = None
    instructions: list[dis.Instruction] = None
    md_env: dict = None


class PidgyCompiler:
    def ast_parse(self, source, filename="<unknown>", symbol="exec"):
        get_ipython().current_execution.nodes = super().ast_parse(source, filename, symbol)
        return get_ipython().current_execution.nodes


def set_compiler_class(shell):
    if not issubclass(shell.compiler_class, PidgyCompiler):
        name = PidgyCompiler.__name__.replace("Compiler", shell.compiler_class.__name__)
        shell.compiler_class = type(name, (PidgyCompiler, shell.compiler_class), {})


def set_current_execution(shell):
    if shell.has_trait("current_execution"):
        id = get_cell_id(shell)
        if shell.current_execution:
            if shell.current_execution.id == id:
                return
        shell.current_execution = Execution(id)


def update_current_execution(lines):
    shell = get_ipython()
    set_current_execution(shell)
    return lines


def load_ipython_extension(shell):
    from traitlets import Instance

    set_compiler_class(shell)
    shell.compile = shell.compiler_class()
    shell.add_traits(current_execution=Instance(Execution, (None,)))
    with suppress(ValueError):
        shell.input_transformer_manager.cleanup_transforms.remove(update_current_execution)
    shell.input_transformer_manager.cleanup_transforms.insert(0, update_current_execution)


def unload_ipython_extension(shell):
    with suppress(ValueError):
        shell.input_transformer_manager.cleanup_transforms.remove(update_current_execution)

    shell.input_transformer_manager.cleanup_transforms.remove(update_current_execution, None)
    if issubclass(shell.compiler_class, PidgyCompiler):
        shell.compiler_class = shell.compiler_class.__mro__[2]
        shell.compile = shell.compiler_class()



"""dodo.py"""
import json
import os
import shutil
import subprocess
import sys
from hashlib import sha256
from pathlib import Path
from re import sub

import doit.tools

PY = Path(sys.executable)
DOIT_CONFIG = dict(verbosity=2)
DODO = Path(__file__)
HERE = DODO.parent
CONF = HERE / "conf.py"
DIST = HERE / "dist"
DOCS = HERE / "docs"
LITE = HERE / "lite"
LICENSE = HERE / "LICENSE"
EXT = LITE / "jupyterlite-pidgy"
LITE_REQS = LITE / "requirements.txt"
LITE_PYPI = LITE / "pypi"
LITE_PYPI_SHA256SUMS = LITE_PYPI / "SHA256SUMS"
LITE_SHA256SUMS = LITE / "_/_/SHA256SUMS"
EXT_LICENSE = EXT / LICENSE.name
EXT_SRC_PKG = EXT / "package.json"
EXT_SRC_PKG_DATA = json.load(EXT_SRC_PKG.open())
EXT_DIST_PKG = EXT / "py_src/jupyterlite_pidgy/labextension/package.json"
EXT_DIST = EXT / "dist"
EXT_WHL_NAME = f"""jupyterlite_pidgy-{EXT_SRC_PKG_DATA["version"]}-py3-none-any.whl"""
EXT_WHL = EXT_DIST / EXT_WHL_NAME
EXT_ICON = EXT / "style/pidgy.png"
SHA256SUMS = DIST / "SHA256SUMS"
KERNEL_DATA = HERE / "src/kernelspec"
KERNEL_ICON = KERNEL_DATA / "logo-64x64.png"
WHL_PY = [
    p
    for p in [
        HERE / "pyproject.toml",
        *Path("src").rglob("*.py"),
    ]
    if p.name != "_version.py"
]
WHL_MD = [HERE / "README.md", Path("src/pidgy/README.md")]
WHL_DEPS = [
    HERE / "pyproject.toml",
    *KERNEL_DATA.rglob("*"),
    *WHL_MD,
    *WHL_PY,
]
ALL_PY = [DODO, *WHL_PY]
ALL_MD = [*WHL_MD]
ALL_JSON = [
    *LITE.glob("*.json"),
    *LITE.glob("*/*.json"),
    *[p for p in WHL_DEPS if p.suffix == "json"],
]
SOURCE_DATE_EPOCH = (
    subprocess.check_output(["git", "log", "-1", "--format=%ct"]).decode("utf-8").strip()
)
os.environ.update(SOURCE_DATE_EPOCH=SOURCE_DATE_EPOCH)


def task_lint():
    """apply source formatting"""
    yield dict(name="black", file_dep=ALL_PY, actions=[["black", *ALL_PY]])

    for a_json in ALL_JSON:
        yield dict(
            name=f"json:{a_json.relative_to(HERE)}",
            file_dep=[a_json],
            actions=[
                lambda: a_json.write_text(
                    json.dumps(
                        json.loads(a_json.read_text(encoding="utf-8")),
                        indent=2,
                        sort_keys=True,
                    )
                    + "\n",
                    encoding="utf-8",
                )
                > 0
            ],
        )


def build_hashfile(hashfile, get_deps):
    deps = sorted(get_deps())

    if not deps:
        raise Exception(f"no files found to hash for {hashfile}")
    lines = []

    for p in deps:
        lines += ["  ".join([sha256(p.read_bytes()).hexdigest(), p.name])]

    output = "\n".join(lines)
    print(output)
    hashfile.write_text(output)


def task_dist():
    """build distributions"""
    yield dict(
        name="pidgy",
        doc="build pidgy distributions",
        file_dep=WHL_DEPS,
        actions=[
            lambda: [shutil.rmtree(DIST) if DIST.is_dir() else None, None][-1],
            [PY, "-m", "hatch", "build", "-t", "sdist"],
            [PY, "-m", "pip", "wheel", "-w", DIST, "--no-deps", "."],
            (
                build_hashfile,
                [
                    SHA256SUMS,
                    lambda: [*DIST.glob("pidgy*.whl"), *DIST.glob("pidgy*.tar.gz")],
                ],
            ),
        ],
        targets=[SHA256SUMS],
    )


@doit.create_after("dist")
def task_labext():
    pkg = EXT / "package.json"
    lock = EXT / "yarn.lock"
    integrity = EXT / "node_modules/.yarn-integrity"
    ts_src = [*(EXT / "src").rglob("*.ts")]
    ts_buildinfo = EXT / "tsconfig.tsbuildinfo"
    pypi = EXT / "pypi/pidgy"
    pypi_ts = EXT / "src/_pypi.ts"

    def _do(*args, **kwargs):
        cwd = str(kwargs.pop("cwd", EXT))
        return doit.tools.CmdAction([*args], **kwargs, cwd=cwd, shell=False)

    def _copy_static():
        src = sorted(DIST.glob("*.whl"))[-1]
        dest = pypi / src.name
        if pypi.exists():
            shutil.rmtree(pypi)
        copy_one(src, dest)
        ctx = "!!file-loader?name=pypi/pidgy/[name].[ext]&context=.!../pypi/pidgy"
        lines = [
            f"""export * as allJSONUrl from '{ctx}/all.json';""",
            f"""export * as pidgyWheelUrl from '{ctx}/{dest.name}';""",
        ]
        pypi_ts.write_text("\n".join(lines))

        copy_one(KERNEL_ICON, EXT_ICON)
        copy_one(LICENSE, EXT_LICENSE)

    yield dict(
        name="yarn",
        file_dep=[pkg] + ([lock] if lock.exists() else []),
        targets=[integrity],
        actions=[_do("jlpm", "--prefer-offline", "--ignore-optional")],
    )

    yield dict(
        name="copy:static",
        file_dep=[*DIST.glob("*.whl"), KERNEL_ICON, LICENSE],
        targets=[pypi / "all.json", pypi_ts, EXT_ICON, EXT_LICENSE],
        actions=[_copy_static, _do("jupyter", "lite", "pip", "index", pypi)],
    )

    yield dict(
        name="lint",
        file_dep=[*ts_src],
        actions=[
            _do("jlpm", "eslint"),
        ],
    )

    yield dict(
        name="build:ts",
        file_dep=[*ts_src, pkg, integrity, pypi_ts],
        targets=[ts_buildinfo],
        actions=[_do("jlpm", "build:lib")],
    )

    yield dict(
        name="build:ext",
        file_dep=[ts_buildinfo, integrity, pkg, pypi_ts, pypi / "all.json"],
        targets=[EXT_DIST_PKG],
        actions=[_do("jlpm", "build:labextension")],
    )

    yield dict(
        name="wheel:ext",
        file_dep=[EXT_DIST_PKG, EXT_LICENSE, SHA256SUMS],
        actions=[_do(PY, "-m", "pip", "wheel", "--no-deps", "-w", EXT_DIST, ".")],
        targets=[EXT_WHL],
    )


@doit.create_after("labext")
def task_lite():
    """build jupyterlite site and pre-requisites"""

    def _wheels():
        if LITE_PYPI_SHA256SUMS.exists():
            LITE_PYPI_SHA256SUMS.unlink()

        wheel = sorted(DIST.glob("pidgy-*.whl"))[-1]

        for old_wheel in LITE_PYPI.glob("pidgy-*.whl"):
            old_wheel.unlink()

        rc = subprocess.check_call(
            [PY, "-m", "pip", "wheel", "--prefer-binary", wheel, "-r", LITE_REQS],
            cwd=str(LITE_PYPI),
        )

        if rc == 0:
            build_hashfile(LITE_PYPI_SHA256SUMS, lambda: [*LITE_PYPI.glob("*py3-none-any.whl")])
            return True

        return False

    def _clean():
        """remove conflicting packages that get patched"""
        for x in ["tornado", "ipython", "ipykernel", "psutil", "pyzmq"]:
            for file in LITE_PYPI.glob(F"{x}*.whl"):
                file.unlink()

    yield dict(
        name="wheels",
        file_dep=[SHA256SUMS, LITE_REQS, EXT_WHL],
        actions=[(doit.tools.create_folder, [LITE / "pypi"]), _wheels, _clean],
    )

    def _lite(args, extra_args=None):
        lite_cmd = [PY, "-m", "jupyter", "lite"]
        lite_args = ["--debug", "--LiteBuildConfig.federated_extensions", EXT_WHL]
        extra_args = extra_args or []
        return doit.tools.CmdAction(
            [*lite_cmd, *args, *lite_args, *extra_args],
            cwd=str(LITE),
            shell=False,
        )

    yield dict(
        name="build",
        file_dep=[
            SHA256SUMS,
            EXT_WHL,
            *LITE.glob("*.json"),
            *(LITE / "retro").glob("*.json"),
            *DOCS.rglob("*.ipynb"),
        ],
        actions=[
            _lite(["build"]),
            _lite(["doit"], ["--", "pre_archive:report:SHA256SUMS"]),
        ],
        targets=[LITE_SHA256SUMS],
    )


@doit.create_after("lite")
def task_docs():
    def post():
        CONF.write_text(
            "\n".join(
                [
                    "import subprocess",
                    """subprocess.check_call(["doit", "lite"])""",
                    sub(
                        r'external_toc_path = "\S+_toc.yml"',
                        r'external_toc_path = "_toc.yml"',
                        CONF.read_text(),
                    ),
                ]
            )
        )

    yield dict(
        name="sphinx-config",
        file_dep=["_toc.yml", "_config.yml"],
        actions=[
            "jb config sphinx .",
            post,
            "black conf.py",
        ],
        targets=[CONF],
    )

    yield dict(
        name="sphinx-build",
        file_dep=[
            CONF,
            *DOCS.rglob("*.ipynb"),
            *(HERE / "_templates").glob("*.html"),
            LITE_SHA256SUMS,
        ],
        actions=["sphinx-build . _build/html %(pos)s"],
        pos_arg="pos",
        targets=[HERE / "_build/html/.buildinfo"],
    )


def copy_one(src, dest):
    if not src.exists():
        return False
    if not dest.parent.exists():
        dest.parent.mkdir(parents=True)
    if dest.exists():
        if dest.is_dir():
            shutil.rmtree(dest)
        else:
            dest.unlink()
    if src.is_dir():
        shutil.copytree(src, dest)
    else:
        shutil.copy2(src, dest)



from collections import defaultdict
from re import compile
from asyncio import ensure_future

from jinja2.meta import find_undeclared_variables
from markdown_it import MarkdownIt
from traitlets import Bool, Dict, Instance, Type, HasTraits, observe, Enum

from .displays import IPythonMarkdown, TemplateDisplay, is_widget
from .environment import IPythonTemplate
from . import get_ipython

CELL_MAGIC = compile("^\s*%{2}\S")
NO_SHOW = compile(r"^\s*\r?\n")

URL = compile("^(http[s]|file)://")
REFS = compile("\!?(\[.*\])?(\[(?P<a>[^\[]+)\])")


class Weave(HasTraits):
    enabled = Bool(True).tag(config=True)
    displays = Dict()
    shell = Instance("IPython.InteractiveShell", ())
    prior = Dict()  # prior value to compare against when reacting to updates
    template_type = Enum(["markdown", "html", "widget"]).tag(config=True)
    template_cls = Type(IPythonMarkdown, TemplateDisplay)
    markdown_renderer = Instance(MarkdownIt, args=())
    reactive = Bool(True).tag(config=True)
    use_async = Bool(True).tag(config=True)
    widgets = Dict()

    @observe("template_type")
    def _set_template_cls(self, value):
        from .displays import IPythonMarkdown, IPyWidgetsHtml, IPythonHtml

        self.template_cls = dict(markdown=IPythonMarkdown, widget=IPyWidgetsHtml, html=IPythonHtml)[
            self.template_type
        ]

    def get_template_vars(self, body):
        return find_undeclared_variables(self.shell.environment.parse(body))

    async def aweave_cell(self, body):
        template = self.shell.environment.from_string(
            body, None, IPythonTemplate, enable_async=True
        )
        return self.template_cls(
            body=body,
            template=template,
            vars=self.get_template_vars(body),
            markdown_renderer=self.markdown_renderer,
            use_async=self.use_async,
        )

    def filter_meta_tokens(self, body):
        if self.shell.has_trait("current_execution") and self.shell.current_execution.tokens:
            for token in self.shell.current_execution.tokens:
                if token.type in {"front_matter", "shebang"}:
                    continue
                break
            return "".join(body.splitlines(1)[token.map[0] :])
        return body

    def _append_references(self, body):
        # we can't reuse markdown references across cells without
        # intervention. this function builds a body of references
        # that might be undefined in the markdown block, but
        # defined in a prior cell.

        from markdown_it.common.utils import normalizeReference

        defs = ""
        for t in self.shell.current_execution.tokens:
            if t.type == "definition":
                defs += f"* [{t.meta['label']}]\n"
                continue
            break
        else:
            return "\n" * 3 + defs

        extras = {}

        all = self.shell.tangle.env.get("references", {})
        this = self.shell.current_execution.md_env.get("references", {})
        for m in REFS.finditer(body):
            r = normalizeReference(m.group("a"))
            if r in extras:
                continue
            if r in this:
                continue
            if r in all:
                ref = all[r]
                extra = f"[{r}]: " + ref["href"]
                if ref.get("title"):
                    extra += f' "{ref["title"]}"'
                extra += "\n"
                extras[r] = extra

        if extras:
            return "\n" * 3 + "\n".join(extras.values())
        return ""

    def weave_cell(self, body):
        body = self.filter_meta_tokens(body)
        if self.shell.tangle.env.get("references"):
            body += self._append_references(body)
        template = self.shell.environment.from_string(body, None, IPythonTemplate)
        vars = find_undeclared_variables(self.shell.environment.parse(body))
        return self.template_cls(
            body=body, template=template, vars=vars, markdown_renderer=self.markdown_renderer
        )

    def get_value(self, key, raw=True):
        """get a value in a namespace that is potentially a widgets."""
        value = self.shell.user_ns.get(key)

        if raw:
            return value

        if is_widget(value):
            return value.value

        return value

    def get_id(self):
        return self.shell.kernel.get_parent().get("metadata", {}).get("cellId")

    def get_vars(self):
        data = defaultdict(list)
        for d in self.displays.values():
            for k in d.vars:
                data[k].append(d)

        return data

    def link_widgets(self):
        # collect all the widgets into a dict
        # link widgets to a display
        displays_by_key = self.get_vars()
        olds = list()
        for key, displays in displays_by_key.items():
            value = self.get_value(key)

            if is_widget(value):
                if self.widgets.setdefault(key, value) is not value:
                    olds.append(self.widgets[key])
                    self.widgets[key] = value

                for display in displays:
                    value.observe(display.observe, "value")

        for old in olds:
            old.close()

    def weave(self, body):
        id = self.get_id()
        self.displays[id] = self.weave_cell(body)
        return self.displays[id]

    def post_run_cell(self, result):
        from IPython.display import display, Markdown

        if not self.enabled:
            return

        id = self.get_id()
        if result.error_in_exec or result.error_before_exec:
            pass  # don't do anything when there are errors
        elif CELL_MAGIC.match(result.info.raw_cell):
            pass
        elif NO_SHOW.match(result.info.raw_cell):
            display(Markdown(f"<div hidden>\n\n{result.info.raw_cell}\n\n</div>"))
        else:
            return display(self.weave(result.info.raw_cell))

        self.displays.pop(id, None)

    def del_displays(self, metadata):
        for id in metadata.get("deletedCells", []):
            # clear and deleted displays
            if id in self.displays:
                del self.displays[id]

    def pre_execute(self):
        if not self.enabled:
            return

        metadata = self.shell.kernel.get_parent().get("metadata", {})
        self.del_displays(metadata)

        vars = set()
        for id, disp in self.displays.items():
            if disp.vars:
                vars.update(disp.vars)
        # collect the state of any reactive or tracked variables before execution
        self.prior.update(zip(vars, map(self.get_value, vars)))

    def post_execute(self, force=False):
        if force or (self.enabled and self.reactive):
            changed = set()

            for k, v in self.prior.items():
                y = self.get_value(k)
                if y is not v:
                    changed.add(k)

            if self.reactive:
                for disp in self.displays.values():
                    if changed.intersection(disp.vars):
                        ensure_future(disp.aupdate())

            self.link_widgets()

    def update(self):
        self.post_execute(True)


def _add_weave_trait(shell):
    from . import environment

    environment.load_ipython_extension(shell)

    if not shell.has_trait("weave"):
        shell.add_traits(
            weave=Instance(
                Weave, kw=dict(shell=shell, markdown_renderer=shell.tangle.parser), allow_none=True
            ).tag(config=True)
        )


def _add_interactivity(shell):
    shell.events.register("pre_execute", shell.weave.pre_execute)
    shell.events.register("post_execute", shell.weave.post_execute)


def _rm_interactivity(shell):
    for e in ["pre_execute", "post_run_cell", "post_execute"]:
        try:
            shell.events.unregister(e, getattr(shell.weave, e))
        except ValueError:
            pass


def load_ipython_extension(shell):
    _add_weave_trait(shell)
    _add_interactivity(shell)
    shell.events.register("post_run_cell", shell.weave.post_run_cell)

    if is_voila():
        from .displays import IPyWidgetsHtml

        # when voila runs a pidgy application we use the ipywidgets template
        shell.weave.template_cls = IPyWidgetsHtml


def unload_ipython_extension(shell):
    if shell.has_trait("weave"):
        _rm_interactivity(shell)


def is_voila():
    """determine if voila is running from the environment variables."""
    from os import environ

    return any(x for x in environ if "voila" in x.lower())



"""pidgy literate computing frame"""
from ._version import __version__

IS_IPY = bool(
    getattr(
        __import__("sys").modules.get("IPython", __import__("sys")),
        "get_ipython",
        lambda: None,
    )()
)


def get_ipython():
    import IPython

    shell = IPython.get_ipython()
    if shell is None:
        shell = IPython.InteractiveShell()
        IPython.get_ipython = shell.get_ipython
    return shell


def get_cell_id(shell=None):
    return (shell or get_ipython()).kernel.get_parent().get("metadata", {}).get("cellId")


def load_ipython_extension(shell):
    shell.enable_html_pager = True

    shell.user_ns.setdefault("shell", shell)
    for e in ("tangle", "weave", "extras", "inspect", "magics"):
        shell.run_line_magic("load_ext", F"pidgy.{e}")


def unload_ipython_extension(shell):
    for e in ("tangle", "weave", "extras", "inspect", "magics"):
        shell.run_line_magic("unload_ext", F"pidgy.{e}")



from midgy.loader import Markdown

from .extras import ReturnDisplay

class Pidgy(Markdown):
    """an importnb extension for pidgy documents"""

    extensions = ".py.md", ".md", ".md.ipynb"

    def visit(self, node):
        return super().visit(ReturnDisplay().visit(node))



"""jupyterlite-pidgy setup"""
import json
from pathlib import Path

HERE = Path(__file__).parent.resolve()
EXT = HERE / "py_src/jupyterlite_pidgy/labextension"
PKG = EXT / "package.json"

__js__ = json.load(PKG.open())
SHARE = f"""share/jupyter/labextensions/{__js__["name"]}"""

setup_args = dict(
    version=__js__["version"],
    data_files=[
        (SHARE, ["install.json"])
    ] + [
        (f"""{SHARE}/{p.parent.relative_to(EXT).as_posix()}""", [
            str(p.relative_to(HERE).as_posix())
        ])
        for p in EXT.rglob("*") if not p.is_dir()
    ]
)

if __name__ == "__main__":
    import setuptools
    setuptools.setup(**setup_args)



"""a modified ipython inspector that gives keystroke resolution rendering."""

from . import get_ipython
from functools import lru_cache, partial
from io import StringIO
import sys, types
from traitlets import Any
from pathlib import Path

INSPECT_METHOD = ["do_inspect", "inspect"][sys.platform == "emscripten"]
from itertools import chain
from contextlib import suppress

DIR = Path(__file__).parent
HELP = DIR / "help"


# the do_inspect method overrides kernel inspection in lab and lite.
def do_inspect(self, cell, cursor_pos, detail_level=1, omit_sections=(), *, cache={}):
    return get_ipython().markdown_inspector.inspect(cell, cursor_pos)


def load_ipython_extension(shell):
    if not shell.kernel.has_trait("original_inspector"):
        shell.kernel.add_traits(original_inspector=Any(getattr(shell.kernel, INSPECT_METHOD)))
    if not shell.has_trait("markdown_inspector"):
        shell.add_traits(markdown_inspector=Any(MarkdownInspector()))

    # we can just modify the shell inspector because that is responsible for discovery
    # objects in the run time. we need to intercept the inspection before it passe sinfo to ipython
    setattr(shell.kernel, INSPECT_METHOD, types.MethodType(do_inspect, shell.kernel))


def unload_ipython_extension(shell):
    if shell.kernel.has_trait("original_inspector"):
        setattr(shell.kernel, INSPECT_METHOD, shell.kernel.original_inspector)


class MarkdownInspector:
    def prepare(self, body):
        return dict(found=True, status="ok", data={"text/markdown": body}, metadata={})

    def inspect(self, code, cursor_pos=0, cache={}):
        tokens = cache.get(code)
        line, offset = lineno_at_cursor(code, cursor_pos)

        if tokens is None:
            cache.clear()
            tokens = cache.setdefault(code, get_ipython().tangle.parse(code))

        where = get_md_token(tokens, line, offset)
        location = f"`L{line+1}:{offset+1}@" + "/".join(x.type for x in where) + "`" + "\n" * 3
        if where:
            explicit = self.visit(where[-1], code, line, offset, cursor_pos)
            if explicit:
                if isinstance(explicit, dict):
                    shell = get_ipython()
                    if "text/html" in explicit.get("data", {}):
                        explicit["data"]["text/html"] = (
                            shell.weave.markdown_renderer.render(location)
                            + explicit["data"]["text/html"]
                        )
                    return explicit
                return self.prepare(location + explicit)
        if code.strip():
            shell = get_ipython()
            out = location + code
            if shell.tangle.env.get("references"):
                # append any prior markdown references
                out += shell.weave._append_references(code)
            return self.prepare(out)
        return self.prepare(get_pidgy_help())

    def visit(self, node, code, line, offset, cursor_pos):
        with suppress(AttributeError):
            return getattr(self, f"visit_{node.type}")(node, code, line, offset, cursor_pos)

    def visit_fence(self, node, code, line, offset, cursor_pos):
        if node.info:
            if node.info in get_ipython().tangle.include_code_fences:
                return self.visit_code_block(node, code, line, offset, cursor_pos)
            with suppress(AttributeError):
                return getattr(self, f"fence_{node.info}")(node, code, line, offset, cursor_pos)

    def fence_mermaid(self, node, code, line, offset, cursor_pos):
        if line == node.map[0]:
            return get_mermaid_help()
        return self.get_sliced_code(node, code)

    def visit_code_block(self, node, code, line, offset, cursor_pos):
        detail = 0
        if code[cursor_pos - 1] == "?":
            detail += 1
        data = get_ipython().kernel.original_inspector(code, cursor_pos, detail_level=detail)
        if data["found"]:
            return data

    def get_sliced_code(self, node, code):
        body = StringIO()
        for i, line in enumerate(StringIO(code)):
            if node.map[0] <= i < node.map[1]:
                body.write(line)
            if i > node.map[1]:
                break
        return body.getvalue()


@lru_cache
def get_help(name):
    return (DIR / "help" / f"{name}.md").read_text()


get_mermaid_help = partial(get_help, "mermaid")
get_pidgy_help = partial(get_help, "pidgy")


def lineno_at_cursor(cell, cursor_pos=0):
    i = offset = 0
    for i, line in enumerate(cell.splitlines(True)):
        next_offset = offset + len(line)
        if not line.endswith("\n"):
            next_offset += 1
        if next_offset > cursor_pos:
            break
        offset = next_offset
    col = cursor_pos - offset
    return i, col


def get_md_token(tokens, line, offset):
    """return the md token stack corresponding to the cursor position."""
    where = []
    for node in tokens:
        if node.type == "root":
            continue
        if node.map:
            if line < node.map[0]:
                break
            elif node.map[0] <= line < node.map[1]:
                where.append(node)
    return where



"""a mkdocs plugin for pidgy

pidgy is designed to decouple input from output. input is a template for the output.
as a result we want to suppress the inputs, but still make that accessible for an eager reader.

mkdocs is designed for markdown so convert notebooks directly to markdown 
and try mkdocs into things notebooks are standard content"""

import nbconvert, re
import functools, re
import mkdocs
from pathlib import Path

PIDGY = re.compile(r"\s*%(re)?load_ext pidgy[^\.]?")
# https://tonyfast.github.io/tonyfast/xxii/2022-12-31-markdownish-notebook.html#generating-the-template

HEAD = """{% from 'base/jupyter_widgets.html.j2' import jupyter_widgets %}
<script src="{{ resources.require_js_url }}"></script>
{{ jupyter_widgets(resources.jupyter_widgets_base_url, resources.html_manager_semver_range, resources.widget_renderer_url) }}
"""
HERE = Path(__file__).parent

mkdocs.utils.markdown_extensions += (".ipynb",)
# dont tell nobody, but now mkdocs things notebooks are valid content


class PidgyExporter(nbconvert.exporters.HTMLExporter):
    def from_notebook_node(self, nb, resources=None, **kw):
        resources = self._init_resources(dict(is_pidgy=is_pidgy(nb)))
        return super().from_notebook_node(nb, resources, **kw)


class Notebooks(mkdocs.plugins.BasePlugin):
    exporter_cls = PidgyExporter
    config_scheme = (
        # ('foo', mkdocs.config.config_options.Type(str, default='a default value')),
    )

    @functools.lru_cache
    def get_exporter(self, key="mkdocs", **kw):
        kw.setdefault("template_file", key)
        exporter = self.exporter_cls(**kw)
        exporter.environment.filters.setdefault("attachment", replace_attachments)
        from jinja2 import DictLoader

        for loader in exporter.environment.loader.loaders:
            if isinstance(loader, DictLoader):
                loader.mapping[key] = (HERE / "templates" / "ipynb.md.j2").read_text()
                loader.mapping["HEAD"] = HEAD
                break
        return exporter

    def on_page_read_source(self, page, config):
        import nbformat.v4
        from yaml import safe_dump

        if page.file.is_modified():
            if page.file.src_path.endswith((".ipynb",)):
                body = Path(page.file.abs_src_path).read_text()
                nb = nbformat.v4.reads(body)
                exporter = self.get_exporter()
                    
                return "\n\n".join(
                    (
                        "---",
                        safe_dump(_node_to_py(nb.metadata), default_flow_style=True),
                        "---",  # add metadata as front matter
                        "", exporter.from_notebook_node(nb)[0],
                    )
                ).lstrip()

    def on_post_page(self, output, page, config):
        if '<script type="application/vnd.jupyter.widget-state+json">' in output:

            left, sep, right = output.partition("</head")
            exporter = self.get_exporter()
            return (
                left
                + self.get_exporter()
                .environment.get_template("HEAD")
                .render(
                    resources=dict(
                        jupyter_widgets_base_url=exporter.jupyter_widgets_base_url,
                        html_manager_semver_range=exporter.html_manager_semver_range,
                        widget_renderer_url=exporter.widget_renderer_url,
                        require_js_url=exporter.require_js_url,
                    )
                )
                + sep
                + right
            )


    def on_page_markdown(self, markdown, page, config, files):
        import markdown

        title = markdown.Markdown(extensions=config["markdown_extensions"]).convert(page.title)
        page.title = title[len("<p>") : -len("</p>")].strip()


def is_pidgy(nb):
    for cell in nb["cells"]:
        if cell["cell_type"] == "code":
            if PIDGY.match("".join(cell["source"])):
                return True
    return False


def replace_attachments(cell):
    source = "".join(cell["source"])
    if cell.get("attachments"):
        for k, v in cell["attachments"].items():
            for t, v in v.items():
                source = source.replace("attachment:" + k, "data:" + t + ";base64," + v)
    return source

def _node_to_py(object):
    """convert notebook node to python objects"""
    from nbformat import NotebookNode
    if isinstance(object, NotebookNode):
        return {k: _node_to_py(v) for k, v in object.items()} 
    return object

            


import json
from pathlib import Path

__all__ = ["__version__", "__js__"]
__js__ = json.load(
    (Path(__file__).parent.resolve() / "labextension/package.json").read_bytes()
)
__version__ = __js__["version"]



"""the extras extension adds features that improve interactive computing

1. exposes the `shell` variable for quick access to the interactive instance.
2. each execution the sys modules are populated in the namespace.
  
  this feature minimizes time between computing iterations by bringing more completion sooner.
  in an interactive computing session, the sys modules hold variable names not available to
  the author, but when available help continue flow between executions.

3. IPython display objects are added to the user namespace
4. top-level return statement become IPython.display expressions.
5. it should be easy to copy and paste json for beginners. 
     `true, false and null` are buultins.
"""

import builtins
from ast import Call, Expr, NodeTransformer, Tuple, parse, copy_location
from pathlib import Path
from subprocess import CalledProcessError

import IPython
from . import get_ipython


def sys_modules_are_part_of_ns():
    from sys import modules

    shell = get_ipython()
    for k in modules:
        if k:
            if k.startswith("_"):
                continue
            if "." in k:
                continue
            shell.user_ns.setdefault(k, modules[k])


# happens once
def ipython_displays_are_part_of_ns(shell):
    from IPython import display
    from IPython.display import DisplayObject, IFrame, TextDisplayObject

    for k, v in vars(display).items():
        if isinstance(v, type) and issubclass(v, (DisplayObject, IFrame)):
            if v not in {DisplayObject, TextDisplayObject}:
                if k[0].isupper():
                    shell.user_ns.setdefault(k, v)
    try:
        ipywidgets_displays_are_part_of_ns(shell)
    except ModuleNotFoundError:
        pass


def ipywidgets_displays_are_part_of_ns(shell):
    import ipywidgets
    from ipywidgets import Widget

    for k, v in vars(ipywidgets).items():
        if isinstance(v, type) and issubclass(v, Widget):
            if k[0].isupper():
                shell.user_ns.setdefault(k, v)


def shebang_transform(lines):
    for i, line in enumerate(map(str.strip, lines)):
        if line:
            if line.startswith(("#!",)):
                lines[i] = lines[0].replace("#!", "%%", 1)
            break
    return lines


def shebang_cell_magic(line, body):
    import tempfile

    from IPython import get_ipython

    shell = get_ipython()
    with tempfile.NamedTemporaryFile(
        prefix=f"ipython-{shell.execution_count}", delete=False, suffix=".py"
    ) as file:
        file.write(body.encode())
    try:
        get_ipython().system(line + " " + file.name)
    except CalledProcessError:
        pass
    finally:
        Path(file.name).unlink()


class ReturnDisplay(NodeTransformer):
    """transform a return node into an IPython display expression"""

    REPLACEMENT = parse('__import__("IPython").display.display', mode="eval")

    def visit_FunctionDef(self, node):
        return node

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_Return(self, node):
        return copy_location(
            Expr(
                Call(
                    func=self.REPLACEMENT.body,
                    args=node.value.elts if isinstance(node.value, Tuple) else [node.value],
                    keywords=[],
                )
            ),
            node,
        )


def load_ipython_extension(shell: IPython.InteractiveShell):
    builtins.true, builtins.false, builtins.null = True, False, None
    shell.events.register("pre_execute", sys_modules_are_part_of_ns)
    ipython_displays_are_part_of_ns(shell)
    shell.ast_transformers.append(ReturnDisplay())
    shell.input_transformers_cleanup.append(shebang_transform)
    shell.register_magic_function(shebang_cell_magic, "cell", "/usr/bin/env")


def unload_ipython_extension(shell: IPython.InteractiveShell):
    try:
        shell.events.unregister("pre_execute", sys_modules_are_part_of_ns)
    except ValueError:
        pass
    shell.ast_transformers = [x for x in shell.ast_transformers if not isinstance(x, ReturnDisplay)]
    try:
        shell.input_transformers_cleanup.remove(shebang_transform)
    except ValueError:
        pass



"""tangle.py provides the Markdown to Python translation model.
"""
from dataclasses import dataclass, field
from io import StringIO
from re import MULTILINE, compile
from urllib.parse import urlparse

from midgy.python import Python

from .current import Execution
from . import get_cell_id, get_ipython

# an instance of this class is used to transform markdown to valid python
# in the ipython extension. the python conversion is constrained by being
# a line for line transformation using indent code blocks (not code fences)
# as references for translating the markdown to valid python objects.

IS_MAGIC = compile("^\s*%{2}", MULTILINE)


def set_current_execution(shell):
    if shell.has_trait("pidgy"):
        id = get_cell_id(shell)
        if shell.pidgy.current_execution:
            if shell.pidgy.current_execution.id == id:
                return
        shell.pidgy.current_execution = Execution(id)


def pidgy_render_lines(lines):
    shell = get_ipython()
    set_current_execution(shell)

    lines = "".join(lines)
    if IS_MAGIC.match(lines):
        return lines.splitlines(True)

    tokens = shell.tangle.parse(lines)
    if hasattr(shell, "current_execution"):
        shell.current_execution.tokens = tokens

    shell.current_execution.py = shell.tangle.render_tokens(tokens, src=lines)
    return shell.current_execution.py.splitlines(True)


@dataclass
class IPython(Python):
    """a markdown to python transpiler meant o be used in IPython"""

    parent: object = field(default_factory=get_ipython)
    last_tokens: list = None
    env: dict = field(default_factory=dict)
    URL_PROTOCOLS = "file", "http", "https"
    VALID_URL_LIST_TOKENS = {
        "paragraph_open",
        "paragraph_close",
        "bullet_list_open",
        "bullet_list_close",
        "list_item_open",
        "list_item_close",
    }

    @classmethod
    def get_url_list(cls, tokens):
        urls = []
        for token in tokens:
            if token.type in cls.VALID_URL_LIST_TOKENS:
                continue
            elif token.type == "inline":
                for line in StringIO(token.content):
                    parsed = urlparse(line)
                    if parsed.netloc in cls.URL_PROTOCOLS:
                        urls.append(line.strip())
                        continue
                    break
                else:
                    continue
                break
            break
        else:
            return urls

    def parse(self, src):
        """parse a cell and add information to the current execution."""
        get_ipython().current_execution.md_env = env = dict()
        tokens = super().parse(src, env)
        refs, dups = env.get("references"), env.get("duplicates_refs")
        if refs:
            self.env.setdefault("references", refs).update(refs)
        if dups:
            self.env.setdefault("duplicates_refs", [])
            self.extend(dups)
        return tokens


def _add_tangle_trait(shell):
    if not shell.has_trait("tangle"):
        from traitlets import Instance

        shell.add_traits(tangle=Instance(IPython, ()))
    from . import current

    current.load_ipython_extension(shell)


def load_ipython_extension(shell):

    _add_tangle_trait(shell)
    shell.input_transformer_manager.cleanup_transforms.insert(0, pidgy_render_lines)


def unload_ipython_extension(shell):
    from . import current

    try:
        shell.input_transformer_manager.cleanup_transforms.remove(pidgy_render_lines)
    except ValueError:
        pass



"""pidgy weaves commonmark markdown into rich, reactive displays."""

from re import compile
from typing import ChainMap

from IPython import get_ipython
from jinja2 import Environment, Template, Undefined
from traitlets import Instance

CELL_MAGIC = compile("^\s*%{2}\S")
NO_SHOW = compile(r"^\s*\r?\n")
URL = compile("^(http[s]|file)://")


class IPythonFinalizer:
    """a finalizer that extracts mimetype data from IPython displays"""

    @staticmethod
    def normalize(type, object, metadata) -> str:
        """normalize and object with (mime)type and return a string."""
        # indents can affect the display when mimetypes are
        # converted to markdown dislays 
        if type == "text/html" or "svg" in type:
            object = get_minified(object)

        if type.startswith("image"):
            md = metadata.get(type, {})
            width, height = md.get("width"), md.get("height")
            object = get_decoded(object)
            *_, data = type.partition("/")
            object = f"""<img src="data:image/{data};base64,{object}"/>"""

        return object

    def __new__(cls, object):
        """convert an object into a markdown/html representation"""

        shell = get_ipython()
        datum = shell.display_formatter.format(object)
        data, metadata = datum if isinstance(datum, tuple) else (datum, {})
        key = next(filter(data.__contains__, get_active_types(shell)), str(object))
        if key == "text/plain":
            return str(object)
        return cls.normalize(key, data[key], metadata)


class IPythonEnvironment(Environment):
    """a jinja2 environment available in the ipython shell"""

    def init_filters(self):
        from . import filters

        self.filters.update(
            (k, v) for k, v in vars(filters).items() if k[0].isalpha() and callable(v)
        )

    def __init__(self, *args, **kwargs):
        from jinja2 import ChoiceLoader, DictLoader, FileSystemLoader

        kwargs.setdefault("loader", ChoiceLoader([DictLoader({}), FileSystemLoader(".")]))
        kwargs.setdefault("finalize", IPythonFinalizer)
        kwargs.setdefault("undefined", Undefined)
        kwargs.setdefault("enable_async", True)
        super().__init__(*args, **kwargs)
        self.init_filters()


class IPythonTemplate(Template):
    """an ipython shell and namespace aware jinja template"""
    def ns(self, *args, **kwargs):
        import builtins

        ns = get_ipython()
        if ns:
            return ChainMap(kwargs, ns.user_ns, vars(builtins))
        return {}

    def render(self, *args, **kwargs):
        return super().render(self.ns(*args, **kwargs))

    async def render_async(self, *args, **kwargs):
        return await super().render_async(self.ns(*args, **kwargs))

    async def generate_async(self, *args, **kwargs):
        async for x in super().generate_async(self.ns(*args, **kwargs)):
            yield x


class Undefined(Undefined):
    def __str__(self, *args, **kwargs):
        # log that the template failed
        if self._undefined_obj:
            return f"`{self._undefined_name} of {self._undefined_obj} is undefined`"
        return f"`{self._undefined_name} is undefined`"


def get_active_types(shell=None):
    """get the active types in the current IPython shell.
    we ignore latex, but i forget why."""
    shell = shell or get_ipython()
    if shell:
        object = list(shell.display_formatter.active_types)
        object.insert(object.index("text/html"), object.pop(object.index("text/latex")))
        return reversed(object)
    return []


def get_minified(x):
    return x


def get_decoded(object):
    if isinstance(object, bytes):
        from base64 import b64encode

        object = b64encode(object).decode("utf-8")
    return object


def get_environment(reuse=True, _cache={}, **kwargs):
    # use this function to avoid repeat environment instantiation
    shell = get_ipython()
    if shell:
        try:
            return shell.environment
        except AttributeError:
            pass
    if reuse:
        if _cache:
            return _cache[True]
        _cache[True] = IPythonEnvironment(**kwargs)
        return _cache[True]
    return IPythonEnvironment(**kwargs)


def load_ipython_extension(shell):
    if not shell.has_trait("environment"):
        shell.add_traits(environment=Instance(IPythonEnvironment, default_value=get_environment()))


def unload_ipython_extension(shell):
    pass



from argparse import ArgumentParser
from sys import argv

from importnb import Notebook

from .tangle import Tangle
from .weave import Weave

parser = ArgumentParser()
parser.add_argument("file")


def main():
    from .loader import Literate

    ns = parser.parse_args(argv[1:])
    module = Literate.load(ns.file)
    from rich import print
    from rich.markdown import Markdown

    print(Markdown(module._repr_markdown_()))


if __name__ == "__main__":
    main()
